/* eccp97a.c - Source for a client to generate distinguished points as part of
   a solution to the problem of determining the private key to a known public
   key in an elliptic curve cryptosystem of prime order ~2^97.
 */

/* Copyright: parts Robert Harley 1997 (Robert.Harley@inria.fr)
   Copyright: parts British Telecommunications plc 1997, 1998
		(adrian.escott@bt-sys.bt.co.uk, john.sager@bt-sys.bt.co.uk,
		jake.hill@bt-sys.bt.co.uk)

   Usage rights statements below. Where usage restrictions differ in the two
   statements the more restrictive applies.
 */

/*
 * Purpose: Some tools for arithmetic on elliptic curves.
 * Copyright: Robert J. Harley, 03-Feb-1997
 * Contact: Robert.Harley@inria.fr
 * Legalese: You can do what you want with this code as long as this
 *           comment stays attached in a prominent position in
 *           human-readable form, any distribution of this or
 *           derived products is made without profit and includes
 *           full source and is subject to the same conditions.
 */

/*
 * Copyright (c) 1997,1998 by British Telecommunications plc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND BRITISH TELECOMMUNICATIONS PLC DISCLAIMS
 * ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL BRITISH TELECOMMUNICATIONS
 * PLC BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
 * ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 */

#include <sys/types.h>
#include <time.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <signal.h>
#include <unistd.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>

/*#define INVTAB*/


#define	SAVFILE	"eccp97a.sav"
#define	INIFILE	"eccp97a.ini"
#define	BUFFILE	"eccp97a.buf"
#define BUFLEN	1024
#define MSGLEN	399

#define	PROXYPORT	8797
#define	OPROXYADDR	0xc1713c25	/* 193.113.60.37 - eccproxy.labs.bt.com */
#define	IPROXYADDR	0x8492c459	/* 132.146.196.89 - raman.alien.bt.co.uk */
#define	SVRADDR		0x8492c436	/* 132.146.196.54 - klingon.alien.bt.co.uk */
#define MAILADDR	"eccp97@zoo.bt.co.uk"
#ifndef VERSION
#define	VERSION		"010"
#endif
#ifndef CLIENT
#define CLIENT		"LNPPCAA"	/* Linux PowerPC with lots of assembly lang. */
#endif

typedef unsigned long u32;
typedef struct {u32 ex, hi, mi, lo;} u128;

char copyright1[] = "Copyright: parts Robert Harley 1997 (Robert.Harley@inria.fr)";
char copyright2[] = "Copyright: parts British Telecommunications plc 1997, 1998"
		" (adrian.escott@bt-sys.bt.co.uk, john.sager@bt-sys.bt.co.uk, jah@alien.bt.co.uk)";

/* variables set from the ini file */
#define PROTO_NONE	0
#define PROTO_TCP	1
#define PROTO_MAIL	2
int protocol = 0; /* 0 = log only, 1 = TCP, 2 = mail */
int proxyport = PROXYPORT;
u32 proxy1addr = SVRADDR;
u32 proxy2addr = IPROXYADDR;
char frombuf[128] = "anon";
char freeform[256] = {'\0'};

int running;		/* set to 0 by a term signal */


#define MODex	0x1UL
#define MODhi	0x6ea1595eUL
#define MODmi	0xd21ae4d8UL
#define MODlo	0xd8420e35UL

#define INVex	0x0UL
#define INVhi	0x606bb011UL
#define INVmi	0x0bc46a65UL
#define INVlo	0xdbcf541dUL

const u128
		mod  = {0x1UL, 0x6ea1595eUL, 0xd21ae4d8UL, 0xd8420e35UL},   /* field prime */
		inv  = {0x0UL, 0x606bb011UL, 0x0bc46a65UL, 0xdbcf541dUL},  /* 1/mod modulo 2^96 */
		code = {0x1UL, 0x5c8e199aUL, 0x19353f4fUL, 0x1bf8cec1UL},    /* 2^96 modulo mod */
		one  = {0x0UL, 0x915ea6a1UL, 0x2de51627UL, 0x27bdf1cbUL},   /* 2^192 modulo mod */
		ord  = {0x1UL, 0x6ea1595eUL, 0xd21ae98fUL, 0xb6cca20dUL}; /*order of group */

u128
	a  = {0x0UL, 0x47370916UL, 0xa603b076UL, 0x57c305c4UL},
	b  = {0x1UL, 0x124df86dUL, 0x04064f50UL, 0x3d9925aFUL},
	x1 = {0x0UL, 0xd5d9e9dfUL, 0xf58a9232UL, 0xa2749ebcUL},
	y1 = {0x1UL, 0x1b34ae5aUL, 0xab7c7ae5UL, 0x5d6abdb5UL},
	x2 = {0x0UL, 0xdf7e84c4UL, 0x2fef50c5UL, 0x316c508aUL},
	y2 = {0x0UL, 0xf259bc58UL, 0x3729da0fUL, 0xe8b97336UL};


/*== Some simple definitions to ease the readability of the code ==*/

/* test equality of two u128 variables */
#define equal_pp(x, y) ((x->ex ^ y->ex) | (x->hi ^ y->hi) | (x->mi ^ y->mi) | (x->lo ^ y->lo)) == 0UL 
 
#define equal_nn(x, y) ((x.ex ^ y.ex) | (x.hi ^ y.hi) | (x.mi ^ y.mi) | (x.lo ^ y.lo)) == 0UL 

/* not equal */

#define not_equ_nn(x, y) (x.ex != y.ex) || (x.hi != y.hi) || (x.mi != y.mi) || (x.lo != y.lo)
   
/* does z = x + y */
#define add3_ppp(z, x, y) {z->lo = x->lo + y->lo; \
	z->mi = z->lo < x->lo;	\
	z->mi += x->mi;			\
	z->hi = z->mi < x->mi;	\
	z->mi += y->mi;			\
	z->hi += z->mi < y->mi;	\
	z->hi += x->hi;			\
	z->ex = z->hi < x->hi;	\
	z->hi += y->hi;			\
	z->ex += z->hi < y->hi;	\
	z->ex += x->ex + y->ex; }

/* does x = x + y */
#define add2_pp(x, y) {x->lo += y->lo; \
	c = x->lo < y->lo;		\
	x->mi += c;				\
	c = x->mi < c;			\
	x->mi += y->mi;			\
	c |= x->mi < y->mi;		\
	x->hi += c;				\
	c = x->hi < c;			\
	x->hi += y->hi;			\
	c |= x->hi < y->hi;		\
	x->ex += c;				\
	x->ex += y->ex; }

#define add2_np(x, y) {x.lo += y->lo; \
	c = x.lo < y->lo;		\
	x.mi += c;				\
	c = x.mi < c;			\
	x.mi += y->mi;			\
	c |= x.mi < y->mi;		\
	x.hi += c;				\
	c = x.hi < c;			\
	x.hi += y->hi;			\
	c |= x.hi < y->hi;		\
	x.ex += c;				\
	x.ex += y->ex; }


#define add2_pn(x, y) {x->lo += y.lo; \
	c = x->lo < y.lo;		\
	x->mi += c;				\
	c = x->mi < c;			\
	x->mi += y.mi;			\
	c |= x->mi < y.mi;		\
	x->hi += c;				\
	c = x->hi < c;			\
	x->hi += y.hi;			\
	c |= x->hi < y.hi;		\
	x->ex += c;				\
	x->ex += y.ex; }

#define add2_nn(x, y) {x.lo += y.lo; \
	c = x.lo < y.lo;		\
	x.mi += c;				\
	c = x.mi < c;			\
	x.mi += y.mi;			\
	c |= x.mi < y.mi;		\
	x.hi += c;				\
	c = x.hi < c;			\
	x.hi += y.hi;			\
	c |= x.hi < y.hi;		\
	x.ex += c;				\
	x.ex += y.ex; }

/* does z = x - y */
#define sub3_pnn(z, x, y) { c = x.lo < y.lo;	\
	z->lo = x.lo - y.lo;				\
	z->mi = x.mi - y.mi;				\
	c1 = (x.mi < y.mi) | (z->mi < c);	\
	z->mi -= c;							\
	z->hi = x.hi - y.hi;				\
	c = (x.hi < y.hi) | (z->hi < c1);	\
	z->hi -= c1;						\
	z->ex = x.ex - y.ex;				\
	z->ex -= c; }

#define sub3_ppp(z, x, y) { c = x->lo < y->lo;	\
	z->lo = x->lo - y->lo;				\
	z->mi = x->mi - y->mi;				\
	c1 = (x->mi < y->mi) | (z->mi < c);	\
	z->mi -= c;							\
	z->hi = x->hi - y->hi;				\
	c = (x->hi < y->hi) | (z->hi < c1);	\
	z->hi -= c1;						\
	z->ex = x->ex - y->ex;				\
	z->ex -= c; }

/* does x = x - y */
#define sub2_pp(x, y) {c = x->lo < y->lo;	\
	x->lo -= y->lo;				\
	c1 = x->mi < c;				\
	x->mi -= c;					\
	c = (x->mi < y->mi) | c1;	\
	x->mi -= y->mi;				\
	c1 = x->hi < c;				\
	x->hi -= c;					\
	c = (x->hi < y->hi) | c1;	\
	x->hi -= y->hi;				\
	x->ex -= c;					\
	x->ex -= y->ex; }

#define sub2_np(x, y) {c = x.lo < y->lo;	\
	x.lo -= y->lo;				\
	c1 = x.mi < c;				\
	x.mi -= c;					\
	c = (x.mi < y->mi) | c1;	\
	x.mi -= y->mi;				\
	c1 = x.hi < c;				\
	x.hi -= c;					\
	c = (x.hi < y->hi) | c1;	\
	x.hi -= y->hi;				\
	x.ex -= c;					\
	x.ex -= y->ex; }

#define sub2_pn(x, y) {c = x->lo < y.lo;	\
	x->lo -= y.lo;				\
	c1 = x->mi < c;				\
	x->mi -= c;					\
	c = (x->mi < y.mi) | c1;	\
	x->mi -= y.mi;				\
	c1 = x->hi < c;				\
	x->hi -= c;					\
	c = (x->hi < y.hi) | c1;	\
	x->hi -= y.hi;				\
	x->ex -= c;					\
	x->ex -= y.ex; }

#define sub2_nn(x, y) {c = x.lo < y.lo;	\
	x.lo -= y.lo;				\
	c1 = x.mi < c;				\
	x.mi -= c;					\
	c = (x.mi < y.mi) | c1;	\
	x.mi -= y.mi;				\
	c1 = x.hi < c;				\
	x.hi -= c;					\
	c = (x.hi < y.hi) | c1;	\
	x.hi -= y.hi;				\
	x.ex -= c;					\
	x.ex -= y.ex; }

/* is x >= y? */
#define greq_pp(x, y) ((x->ex > y->ex) || ((x->ex == y->ex) && (x->hi > y->hi)) \
	|| ((x->ex == y->ex) && (x->hi == y->hi) && (x->mi > y->mi))				\
	|| ((x->ex == y->ex) && (x->hi == y->hi) && (x->mi == y->mi) && (x->lo >= y->lo)))

#define greq_pn(x, y) ((x->ex > y.ex) || ((x->ex == y.ex) && (x->hi > y.hi)) \
	|| ((x->ex == y.ex) && (x->hi == y.hi) && (x->mi > y.mi))				\
	|| ((x->ex == y.ex) && (x->hi == y.hi) && (x->mi == y.mi) && (x->lo >= y.lo)))

#define greq_np(x, y) ((x.ex > y->ex) || ((x.ex == y->ex) && (x.hi > y->hi)) \
	|| ((x.ex == y->ex) && (x.hi == y->hi) && (x.mi > y->mi))				\
	|| ((x.ex == y->ex) && (x.hi == y->hi) && (x.mi == y->mi) && (x.lo >= y->lo)))

#define greq_nn(x, y) ((x.ex > y.ex) || ((x.ex == y.ex) && (x.hi > y.hi)) \
	|| ((x.ex == y.ex) && (x.hi == y.hi) && (x.mi > y.mi))				\
	|| ((x.ex == y.ex) && (x.hi == y.hi) && (x.mi == y.mi) && (x.lo >= y.lo)))

/* is x < y? */
#define less_pp(x, y) ((x->ex < y->ex) || ((x->ex == y->ex) && (x->hi < y->hi)) \
	|| ((x->ex == y->ex) && (x->hi == y->hi) && (x->mi < y->mi))				\
	|| ((x->ex == y->ex) && (x->hi == y->hi) && (x->mi == y->mi) && (x->lo < y->lo)))

#define less_nn(x, y) ((x.ex < y.ex) || ((x.ex == y.ex) && (x.hi < y.hi)) \
	|| ((x.ex == y.ex) && (x.hi == y.hi) && (x.mi < y.mi))				\
	|| ((x.ex == y.ex) && (x.hi == y.hi) && (x.mi == y.mi) && (x.lo < y.lo)))

/* shift left by one */
#define shl_p(x) {x->ex <<= 1;	\
	x->ex |= x->hi>>31;		\
	x->hi <<= 1;			\
	x->hi |= x->mi>>31;		\
	x->mi <<= 1;			\
	x->mi |= x->lo>>31;		\
	x->lo <<= 1; }

#define shl_n(x) {x.ex <<= 1;	\
	x.ex |= x.hi>>31;		\
	x.hi <<= 1;			\
	x.hi |= x.mi>>31;		\
	x.mi <<= 1;			\
	x.mi |= x.lo>>31;		\
	x.lo <<= 1; }						

/* shift right by one */
#define shr_p(x) {x->lo >>= 1;	\
	x->lo |= x->mi<<31;		\
	x->mi >>= 1;			\
	x->mi |= x->hi<<31;		\
	x->hi >>= 1;			\
	x->hi |= x->ex<<31;		\
	x->ex >>= 1; }	

#define shr_n(x) {x.lo >>= 1;	\
	x.lo |= x.mi<<31;		\
	x.mi >>= 1;			\
	x.mi |= x.hi<<31;		\
	x.hi >>= 1;			\
	x.hi |= x.ex<<31;		\
	x.ex >>= 1; }	

/*== Function definitions ==================================================*/

/*== Some simple modular arithmetic ========================================*/

/*-- equal -----------------------------------------------------------------*/

/* Check equality: x == y, 0 <= x, y < some modulus < 2^96. */
static u32 equal(u128 *x, u128 *y)
{
	return equal_pp(x, y);
} /* end equal */


/*-- addMod ----------------------------------------------------------------*/

/* Add: z = x + y modulo odd modulus, 0 <= x, y < modulus < 2^96. */
static u128 addMod(u128 x, u128 y, const u128 mod) 
{ 
	u32 c, c1;
	

	/* adds x and y */
	add2_nn(x, y);

	/* this does x mod m */ 
	if (greq_nn(x, mod))
	{
		sub2_nn(x, mod);
	}

	return x;
} /* end addMod */


/*-- subMod ----------------------------------------------------------------*/

/* Subtract: x - y modulo odd modulus, 0 <= x, y < modulus < 2^96. */
static u128 subMod(u128 x, u128 y, const u128 mod) 
{	
	u32 c, c1;
  
	if (less_nn(x, y))
	  {
		 add2_nn(x, mod);
	  } /* end if */
  
	  sub2_nn(x, y);
  
	  return x;
} /* end subMod */




/*-- doubleMod -------------------------------------------------------------*/

/* Double x modulo odd modulus, 0 <= x < modulus < 2^96. */
static u128 doubleMod(u128 x, const u128 mod) 
{
	u32 c, c1;
	
	shl_n(x);
	
	if (greq_nn(x, mod))
	{
		sub2_nn(x, mod);
	}
	
	return x;
} /* end doubleMod */

/*== Arithmetic for field Z/pZ, p odd, using 2-adic inverse ================*/

#ifdef NOASM
#define CPMUL
#else

#undef CPMUL
#ifdef __GNUC__
#if defined(__i386__)
static inline u32 pmul(u32 *hi, u32 x, u32 y)
{
  u32 t1, t2;

  asm("mull %3" : "=a" (t1), "=d" (t2) : "%0" (x), "rm" (y));
  *hi = t2;
  return t1;
}

#elif defined(__arm6__)
static inline u32 pmul(u32 *hi, u32 x, u32 y)
{
  u32 t1, t2;

  __asm__ ( "umull %1,%0,%2,%3"
          : "=&r" ((u32) (t2)), "=&r" ((u32) (t1))
          : "r" ((u32) (x)), "r" ((u32) (y))
          );
  *hi = t2;
  return t1;
}

#elif defined(_ARCH_PPC)
static inline u32 pmul(u32 *hi, u32 x, u32 y)
{
  u32 t1, t2;

    __asm__ ( "mulhwu %0,%1,%2"
            : "=r" ((u32) t2)
            : "%r" (x), "r" (y)
            );
  *hi = t2;
  return x*y;
}

#elif defined(__sparc__)

static inline u32 pmul(u32 *hi, u32 x, u32 y)
{
  u32 t1, t2;

  __asm__ ( "umul %2,%3,%1;rd %%y,%0"
          : "=r" ((u32) (t2)), "=r" ((u32) (t1))
          : "r" ((u32) (x)), "r" ((u32) (y))
          );
  *hi = t2;
  return x*y;
}

#else
/* put asm in here for other 32->64-bit unsigned multiplies on other architectures */

#define CPMUL
#endif /* architectures */


#else /* !__GNUC__ */
#if defined(__i386__)
/* this must be set up for the appropriate asm syntax for other 386 compilers */
static inline u32 pmul(u32 *hi, u32 x, u32 y)
{
	__asm{
		mov eax,x
		mul y
		mov ebx,hi
		mov [ebx],edx
	}
}
#else
/* non-GCC asm in here for other 32->64-bit unsigned multiplies on other architectures */
#define CPMUL
#endif

#endif /* __GNUC__ */
#endif /* NOASM */

#ifdef CPMUL

/* sorry guys - s n a i l time! If you have to use this, it may be faster to
   set PARAL to a low number, even 1. (PARAL is defined before main(), below)
 */

static inline u32 pmul(u32 *hi, u32 x, u32 y)
{
	u32 xl, xh, yl, yh, t0, t1, t2, t3, t;

	xl = (x & 0xFFFF); xh = (x>>16);
	yl = (y & 0xFFFF); yh = (y>>16);
	t0 = xl * yl;
	t1 = xl * yh;
	t2 = yl * xh;
	t3 = yh * xh;
	t = t1<<16;
	t0 += t;
	t3 += t0 < t;
	t = t2<<16;
	t0 += t;
	t3 += t0 < t;
	t3 += (t1>>16);
	t3 += (t2>>16);

	*hi = t3;

	return t0;
}
#endif

#ifndef PPCFUNCS       /* PowerPC can use inline functions */
static u128 productMod(u128 *x, u128 *y)
{
	u32 ux, uh, um, ul, c, c1, pl1, pl2, pm1;
	u128 u, q, z;

	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6, h7, l7, h8, l8, h9, l9;

		ul = pmul(&um, x->lo, y->lo);
		l2 = pmul(&h2, x->mi, y->lo);
		um += l2; uh = (um < l2);
		l3 = pmul(&h3, x->lo, y->mi);
		um += l3; uh += (um < l3);
		l4 = pmul(&h4, x->mi, y->mi);
		uh += h2; z.lo = (uh < h2);
		l5 = pmul(&h5, x->hi, y->lo);
		uh += h3; z.lo += (uh < h3);
		uh += l4; z.lo += (uh < l4);
		l6 = pmul(&h6, x->lo, y->hi);
		uh += l5; z.lo += (uh < l5);
		uh += l6; z.lo += (uh < l6);
		z.lo += h4; z.mi = (z.lo < h4);
		l7 = pmul(&h7, x->hi, y->mi);
		z.lo += h5; z.mi += (z.lo < h5);
		z.lo += h6; z.mi += (z.lo < h6);
		l8 = pmul(&h8, x->mi, y->hi);
		z.lo += l7; z.mi += (z.lo < l7);
		z.lo += l8; z.mi += (z.lo < l8);
		z.mi += h7; z.hi = (z.mi < h7);
		l9 = pmul(&h9, x->hi, y->hi);
		z.mi += h8; z.hi += (z.mi < h8);
		z.mi += l9; z.hi += (z.mi < l9);
		z.hi += h9; z.ex = (z.hi < h9);

		if (x->ex == 1UL)
			add2_np(z, y);
		if (y->ex == 1UL)
		{
			add2_np(z, x);
			z.ex -= x->ex; /* this removes double added top bit */
		}
	
		ux = (z.lo & 1UL); /* get top bit of lower multiple */
		shr_n(z);            /*shift right for upper multiple */

		/*printf("Pu: %01lx %08lx %08lx %08lx\n\n", ux, uh, um, ul);*/
		/*printf("Pz: %01lx %08lx %08lx %08lx\n\n", z.ex, z.hi, z.mi, z.lo);*/


	}

	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6;

		q.lo = pmul(&q.mi, INVlo, ul);
		l2 = pmul(&h2, INVmi, ul);
		q.mi += l2; q.hi = (q.mi < l2);
		l3 = pmul(&h3, INVlo, um);
		q.mi += l3; q.hi += (q.mi < l3);
		l4 = pmul(&h4, INVmi, um);
		q.hi += h2; q.ex = (q.hi < h2);
		l5 = pmul(&h5, INVhi, ul);
		q.hi += h3; q.ex += (q.hi < h3);
		q.hi += l4; q.ex += (q.hi < l4);
		l6 = pmul(&h6, INVlo, uh);
		q.hi += l5; q.ex += (q.hi < l5);
		q.hi += l6; q.ex += (q.hi < l6);
		q.ex += h4;
		q.ex += h5;
		q.ex += h6;
		q.ex += uh;
		q.ex += um;
		
		q.ex += ux;

		q.ex &= 1UL;
	}

	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6, h7, l7, h8, l8, h9, l9;

		pl1 = pmul(&pl2, MODlo, q.lo); 
		l2 = pmul(&h2, MODmi, q.lo);   
		pl2 += l2; pm1 = (pl2 < l2);
		l3 = pmul(&h3, MODlo, q.mi);   
		pl2 += l3; pm1 += (pl2 < l3);
		l4 = pmul(&h4, MODmi, q.mi);   
		pm1 += h2; u.lo = (pm1 < h2);   
		l5 = pmul(&h5, MODhi, q.lo);
		pm1 += h3; u.lo += (pm1 < h3);
		pm1 += l4; u.lo += (pm1 < l4);
		l6 = pmul(&h6, MODlo, q.hi);
		pm1 += l5; u.lo += (pm1 < l5);
		pm1 += l6; u.lo += (pm1 < l6);
		u.lo += h4; u.mi = (u.lo < h4);
		l7 = pmul(&h7, MODhi, q.mi);
		u.lo += h5; u.mi += (u.lo < h5);
		u.lo += h6; u.mi += (u.lo < h6);
		l8 = pmul(&h8, MODmi, q.hi);
		u.lo += l7; u.mi += (u.lo < l7);
		u.lo += l8; u.mi += (u.lo < l8);
		u.mi += h7; u.hi = (u.mi < h7);
		l9 = pmul(&h9, MODhi, q.hi);
		u.mi += h8; u.hi += (u.mi < h8);
		u.mi += l9; u.hi += (u.mi < l9);
		u.hi += h9; u.ex = (u.hi < h9);

							/* think about these lines */	
		add2_nn(u, q);	/* as we know top bit of mod */
		
		if (q.ex == 1UL)
		{
			add2_nn(u, mod);
			u.ex -= 1; /* this removes double added top bit */
		}

		shr_n(u);            /*shift right for upper multiple */
	}
	 
	if (greq_nn(u, z))
	{
		add2_nn(z, mod)
	} 

	sub2_nn(z, u);	

	return z;  
} /* end productMod */


/*-- squareMod ------------------------------------------------------------*/


static u128 squareMod(u128 *x)
{
	u128 z, q, u, x1;
	u32 ux, uh, um, ul, c, c1, pl1, pl2, pm1;
 
	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6;
		ul = pmul(&um, x->lo, x->lo);
		l2 = pmul(&h2, x->mi, x->lo);
		um += l2; uh = (um < l2);
		um += l2; uh += (um < l2);
		l3 = pmul(&h3, x->mi, x->mi);
		uh += h2; z.lo = (uh < h2);
		uh += h2; z.lo += (uh < h2);
		l4 = pmul(&h4, x->hi, x->lo);
		uh += l3; z.lo += (uh < l3);
		uh += l4; z.lo += (uh < l4);
		uh += l4; z.lo += (uh < l4);
		z.lo += h3; z.mi = (z.lo < h3);
		l5 = pmul(&h5, x->hi, x->mi);
		z.lo += h4; z.mi += (z.lo < h4);
		z.lo += h4; z.mi += (z.lo < h4);
		z.lo += l5; z.mi += (z.lo < l5);
		z.lo += l5; z.mi += (z.lo < l5);
		l6 = pmul(&h6, x->hi, x->hi);
		z.mi += h5; z.hi = (z.mi < h5);
		z.mi += h5; z.hi += (z.mi < h5);
		z.mi += l6; z.hi += (z.mi < l6);
		z.hi += h6;	z.ex = (z.hi < h6);

		if (x->ex == 1UL)
		{
			x1 = *x;
			shl_n(x1)
			add2_nn(z, x1);		
			z.ex -= 1; /* this removes double added top bit */
		}
	
		ux = (z.lo & 1UL); /* get top bit of lower multiple */
		shr_n(z);            /*shift right for upper multiple */ 
	}

	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6;

		q.lo = pmul(&q.mi, INVlo, ul);
		l2 = pmul(&h2, INVmi, ul);
		q.mi += l2; q.hi = (q.mi < l2);
		l3 = pmul(&h3, INVlo, um);
		q.mi += l3; q.hi += (q.mi < l3);
		l4 = pmul(&h4, INVmi, um);
		q.hi += h2; q.ex = (q.hi < h2);
		l5 = pmul(&h5, INVhi, ul);
		q.hi += h3; q.ex += (q.hi < h3);
		q.hi += l4; q.ex += (q.hi < l4);
		l6 = pmul(&h6, INVlo, uh);
		q.hi += l5; q.ex += (q.hi < l5);
		q.hi += l6; q.ex += (q.hi < l6);
		q.ex += h4;
		q.ex += h5;
		q.ex += h6;
		
		q.ex += uh;
		q.ex += um;		
		q.ex += ux;

		q.ex &= 1UL;
	}

	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6, h7, l7, h8, l8, h9, l9;

		pl1 = pmul(&pl2, MODlo, q.lo); 
		l2 = pmul(&h2, MODmi, q.lo);   
		pl2 += l2; pm1 = (pl2 < l2);
		l3 = pmul(&h3, MODlo, q.mi);   
		pl2 += l3; pm1 += (pl2 < l3);
		l4 = pmul(&h4, MODmi, q.mi);   
		pm1 += h2; u.lo = (pm1 < h2);   
		l5 = pmul(&h5, MODhi, q.lo);
		pm1 += h3; u.lo += (pm1 < h3);
		pm1 += l4; u.lo += (pm1 < l4);
		l6 = pmul(&h6, MODlo, q.hi);
		pm1 += l5; u.lo += (pm1 < l5);
		pm1 += l6; u.lo += (pm1 < l6);
		u.lo += h4; u.mi = (u.lo < h4);
		l7 = pmul(&h7, MODhi, q.mi);
		u.lo += h5; u.mi += (u.lo < h5);
		u.lo += h6; u.mi += (u.lo < h6);
		l8 = pmul(&h8, MODmi, q.hi);
		u.lo += l7; u.mi += (u.lo < l7);
		u.lo += l8; u.mi += (u.lo < l8);
		u.mi += h7; u.hi = (u.mi < h7);
		l9 = pmul(&h9, MODhi, q.hi);
		u.mi += h8; u.hi += (u.mi < h8);
		u.mi += l9; u.hi += (u.mi < l9);
		u.hi += h9; u.ex = (u.hi < h9);

							/* think about these lines */	
		add2_nn(u, q);	/* as we know top bit of mod */
		
		if (q.ex == 1UL)
		{
			add2_nn(u, mod);
			u.ex -= 1; /* this removes double added top bit) */
		}

		shr_n(u);            /*shift right for upper multiple */
	}
	 
	if (greq_nn(u, z))
	{
		add2_nn(z, mod)
	} 

	sub2_nn(z, u);

	return z;  
} /* end squareMod */
#else /* PPCFUNCS */
#include "pmod.ppc.gcc"
#define squareMod(x) productMod(x,x)
#endif /* PPCFUNCS */

/* Encode x modulo odd modulus, 0 <= x < modulus < 2^96.
 * Point coordinates are encoded by multiplying by 2^96 mod the prime
 * so that we can multiply modulo the prime quickly using a precomputed
 * 2-adic inverse (as described by Peter Montgomery).
 */
static u128 encode(u128 x) 
{
	u128 z, t;

	t = code;

	 z= productMod(&x, &t);

	return z;
}/* end encode */


#ifndef PPCFUNCS
/*-- decode ----------------------------------------------------------------*/

/* Decode x modulo odd modulus, 0 <= x < modulus < 2^96. */
/* t = HiMul(x*inv, mod);  return t ? mod-t : 0UL; */
static void decode(u128 *v, u128 *x) 
{
	u128 q, u;
	
	{
		u32 h2, l2, h3, l3, h4, l4, h5, l5, h6, l6;

		q.lo = pmul(&q.mi, INVlo, x->lo);
		l2 = pmul(&h2, INVmi, x->lo);
		q.mi += l2; q.hi = (q.mi < l2);
		l3 = pmul(&h3, INVlo, x->mi);
		q.mi += l3; q.hi += (q.mi < l3);
		l4 = pmul(&h4, INVmi, x->mi);
		q.hi += h2; q.ex = (q.hi < h2);
		l5 = pmul(&h5, INVhi, x->lo);
		q.hi += h3; q.ex += (q.hi < h3);
		q.hi += l4; q.ex += (q.hi < l4);
		l6 = pmul(&h6, INVlo, x->hi);
		q.hi += l5; q.ex += (q.hi < l5);
		q.hi += l6; q.ex += (q.hi < l6);
		q.ex += h4;
		q.ex += h5;
		q.ex += h6;
		
		q.ex += x->hi;
		q.ex += x->mi;
		q.ex += x->ex; 

		q.ex &= 1UL;
	}

	{
		u32 c, h2, l2, h3, l3, h4, l4, h5, l5, h6, l6, h7, l7, h8, l8, h9, l9, pl1, pl2, pm1;

		pl1 = pmul(&pl2, MODlo, q.lo); 
		l2 = pmul(&h2, MODmi, q.lo);   
		pl2 += l2; pm1 = (pl2 < l2);
		l3 = pmul(&h3, MODlo, q.mi);   
		pl2 += l3; pm1 += (pl2 < l3);
		l4 = pmul(&h4, MODmi, q.mi);   
		pm1 += h2; u.lo = (pm1 < h2);   
		l5 = pmul(&h5, MODhi, q.lo);
		pm1 += h3; u.lo += (pm1 < h3);
		pm1 += l4; u.lo += (pm1 < l4);
		l6 = pmul(&h6, MODlo, q.hi);
		pm1 += l5; u.lo += (pm1 < l5);
		pm1 += l6; u.lo += (pm1 < l6);
		u.lo += h4; u.mi = (u.lo < h4);
		l7 = pmul(&h7, MODhi, q.mi);
		u.lo += h5; u.mi += (u.lo < h5);
		u.lo += h6; u.mi += (u.lo < h6);
		l8 = pmul(&h8, MODmi, q.hi);
		u.lo += l7; u.mi += (u.lo < l7);
		u.lo += l8; u.mi += (u.lo < l8);
		u.mi += h7; u.hi = (u.mi < h7);
		l9 = pmul(&h9, MODhi, q.hi);
		u.mi += h8; u.hi += (u.mi < h8);
		u.mi += l9; u.hi += (u.mi < l9);
		u.hi += h9; u.ex = (u.hi < h9);

		add2_nn(u, q);

		if (q.ex == 1UL)
		{
			add2_nn(u, mod);
			u.ex -= 1; /* this removes double added top bit) */
		}

		shr_n(u);            /*shift right for upper multiple */
	}

	v->ex = v->hi = v->mi = v->lo = 0UL;

	if (u.ex || u.hi || u.mi || u.lo)
	{ 
		u32 c, c1;
	
		sub3_pnn(v, mod, u)
	}
  
	return;
}/* end decode */
#else /* PPCFUNCS */
#include "decode.ppc.gcc"
#endif /* PPCFUNCS */


/*-- inverseMod ------------------------------------------------------------*/

/* Divide 1 by y modulo odd modulus, 0 <= y < modulus < 2^96, y != 0.
 * Input must be encoded and output is too.
 * Exits with error if inverse doesn't exist (unless NDEBUG is defined).
 */

#define TAB_SIZE (64UL)
#define TAB_MASK (TAB_SIZE-1UL)

static u128 inverseMod(u128 y) 
{
	u32 t, c, c1;
	u128 a, b, v, u;

#ifdef INVTAB
	u32 sh, th;
	static const int tab[TAB_SIZE] = { 
		6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,
		5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0};
#endif

	decode(&y, &y);

	/* Maintain: u.y = a<<t and v.y = -b<<t. */
	
	a = y;
	b = mod;

	t = v.ex = v.hi = v.mi = v.lo = u.ex = u.hi = u.mi = 0UL;

	u.lo = 1UL;

	while (!(a.lo & 1UL))
	{
		shr_n(a)
		t++;
	}

	do
	{
		do	
		{
			sub2_nn(b, a)
			
			add2_nn(v, u)

#ifdef INVTAB
			sh =  tab[(b.lo & 63UL)];
			th = 32UL - sh;

			b.lo >>= sh;
			b.lo |= b.mi<<th;
			b.mi >>= sh;
			b.mi |= b.hi<<th;
			b.hi >>= sh;
			b.hi |= b.ex<<th;
			b.ex >>= sh;

			t += sh;

			u.ex <<= sh;
			u.ex |= u.hi>>th;
			u.hi <<= sh;
			u.hi |= u.mi>>th;
			u.mi <<= sh;
			u.mi |= u.lo>>th;
			u.lo <<= sh;
#endif
			while (!(b.lo & 1UL))
			{
				shr_n(b)

				t++;

				shl_n(u)
			} /* end while */


		} while(less_nn(a, b));

		if (equal_nn(a,b)) break;

		do	
		{
			sub2_nn(a, b)
			
			add2_nn(u, v)

#ifdef INVTAB			
			sh = tab[(a.lo & 63UL)];
			th= 32UL - sh;

			a.lo >>= sh;
			a.lo |= a.mi<<th;
			a.mi >>= sh;
			a.mi |= a.hi<<th;
			a.hi >>= sh;
			a.hi |= a.ex<<th;
			a.ex >>= sh;

			t += sh;

			v.ex <<= sh;
			v.ex |= v.hi>>th;
			v.hi <<= sh;
			v.hi |= v.mi>>th;
			v.mi <<= sh;
			v.mi |= v.lo>>th;
			v.lo <<= sh;
#endif
			while (!(a.lo & 1UL))
			{
				shr_n(a)

				t++;

				shl_n(v)
			} /* end while */

		} while (less_nn(b, a));

	} while (not_equ_nn(a, b));
  
	/* TO DO: consider multiplying by constants from a table here. */
  
	if (t > 97UL) 
	{
		u128 f;

		t -= 97UL;
		f.ex = 0UL;
		f.hi = 0xb750acafUL;
		f.mi = 0x690D726cUL;
		f.lo = 0x6c21071bUL;

		do 
		{
			u32 o;
			
			o = u.lo & 1UL;
			shr_n(u)
      
			if (o) 
			{
				u32 c;

				add2_nn(u, f)
			}      
			--t;
		} while (t);
	} 
	else
	{
		if (t < 97UL) 
		{		
			t = 97UL-t;
		
			do 
			{
				u32 c, c1;
				shl_n(u)
   
				if (greq_nn(u, mod)) 
				{
					sub2_nn(u, mod)
				} /* end if */
      
				--t;
			} while (t);
		} /* end if/else */
	}

	return u;
} /* end inverseMod */

/*-- quotientMod -----------------------------------------------------------*/

/* Divide x by y modulo odd modulus, 0 <= x,y < modulus < 2^96, y != 0. */

static u128 quotientMod(u128 x, u128 y)
{
	u128 z;

	z = inverseMod(y);

	z = productMod(&x, &z);

	return z;
} /* end quotientMod */

/*== Stuff for elliptic curve y^2 = x^3 + a*x + b ==========================*/

/*-- ellipticDouble --------------------------------------------------------*/

/* Given point (x, y) on y^2 = x^3+a*x+b modulo odd modulus, this
 * computes its double by the group law.  Doesn't actually need b.
 * The point at infinity is represented by (0:1:0), with encoded 1.
 * Finite points are represented by (x:y:1), with encoded coordinates x and y.
 * If modulus is not prime, could fail upon finding divisor of it.
 */
static int ellipticDouble(u128 x, u128 y, int z, u128 a, u128 *px2, u128 *py2) 
{  
	u128 den, lam, num, s, t, x2, y2;
	const u128 zero = {0UL, 0UL, 0UL, 0UL};

	if (z == 0 || (y.ex | y.hi | y.mi | y.lo) == 0UL)
	{ 
		*px2 = zero; *py2 = one; return 0; 
	}

	num = squareMod(&x);
	s = doubleMod(num, mod);
	num = addMod(num, s, mod);
	num = addMod(num, a, mod);
	den = doubleMod(y, mod);
	lam = quotientMod(num, den);

	x2 = squareMod(&lam);
	t = doubleMod(x, mod);
  	x2 = subMod(x2, t, mod);
	s = subMod(x, x2, mod);

	y2 = productMod(&lam, &s);  

	y2 = subMod(y2, y, mod);

	*px2 = x2; *py2 = y2; return 1;
} /* end ellipticDouble */

/*-- ellipticSum -----------------------------------------------------------*/

/* Given pts (x1:y1:z1) and (x2:y2:z2) on y^2 = x^3+a*x+b modulo odd modulus,
 * computes their sum by the group law.  Doesn't actually need b.
 * The point at infinity is represented by (0:1:0), with encoded 1.
 * Finite points are represented by (x:y:1), with encoded coordinates x and y.
 * If modulus is not prime, could fail upon finding divisor of it.
 */
static int ellipticSum(u128 x1, u128 y1, int z1, u128 x2, u128 y2, int z2,
					   u128 a, u128 *px3, u128 *py3) 
{
	u128 den, lam, num, s, x3, y3;

	if (z1 == 0) { *px3 = x2; *py3 = y2; return z2; }
  
	if (z2 == 0) { *px3 = x1; *py3 = y1; return z1; }

	if (equal(&x1, &x2)) 
	{
		const u128 zero = {0UL, 0UL, 0UL, 0UL};
		if (equal(&y1, &y2)) 
		{
			return ellipticDouble(x1, y1, z1, a, px3, py3);
		} /* end if */
		*px3 = zero; *py3 = one; return 0;
	} /* end if */
 
	num = subMod(y2, y1, mod);
	den = subMod(x2, x1, mod);
	lam = quotientMod(num, den);

	x3 = squareMod(&lam);
	x3 = subMod(x3, x1, mod);
	x3 = subMod(x3, x2, mod);

	s = subMod(x1, x3, mod);
	y3 = productMod(&lam, &s);
	y3 = subMod(y3, y1, mod);

	*px3 = x3; *py3 = y3; return 1;
} 

/*-- ellipticProduct --------------------------------------------------------*/

/* compute the product of a point (x:y:z) and a number, fac, by repeatedly
   doubling the point and adding it to the running sum if fac contains
   that power of 2. This doesn't need the MS bit of fac to be found
   beforehand.
*/

static int ellipticProduct(u128 x, u128 y, int z, u128 fac, u128 a, 
			   u128 *px2, u128 *py2)
{
	u32 fx, fh, fm, fl, m, z1, z2;
	u128 x1, y1, x2, y2;

	fx = fac.ex; fh = fac.hi; fm = fac.mi; fl = fac.lo;
  
	if (fx == 0UL && fh == 0UL && fl == 0UL && fm == 0UL) 
	{
		px2->ex = px2->hi = px2->mi = px2->lo = 0UL; *py2 = one; return 0;
	} /* end if */

	x2.ex = x2.hi = x2.mi = x2.lo = 0UL; y2 = one; z2 = 0;
	x1 = x; y1 = y; z1 = z;
	m = 1;
  
	while (m) 
	{
		if (fl & m) 
		{
			z2 = ellipticSum(x1, y1, z1, x2, y2, z2, a, &x2, &y2);
		}
    
		z1 = ellipticDouble(x1, y1, z1, a, &x1, &y1);
		m <<= 1;
	}

	m = 1;

	while (m) 
	{
		if (fm & m) 
		{
			z2 = ellipticSum(x1, y1, z1, x2, y2, z2, a, &x2, &y2);
		}
    
		z1 = ellipticDouble(x1, y1, z1, a, &x1, &y1);
		m <<= 1;
	}

	m = 1;

	while (m) 
	{
		if (fh & m) 
		{
			z2 = ellipticSum(x1, y1, z1, x2, y2, z2, a, &x2, &y2);
		}
    
		z1 = ellipticDouble(x1, y1, z1, a, &x1, &y1);
		m <<= 1;
	}
  
	m = 1;
  
		if (fx==1UL)
		{
			z2 = ellipticSum(x1, y1, z1, x2, y2, z2, a, &x2, &y2);	
		}
	
  
	*px2 = x2; *py2 = y2;

	return z2;
}

/* term_sig - signal catcher for SIGTERM */

void
term_sig()
{
  running = 0;
}

/* loadini - get the setup data in the ini file */
int
load_ini()
{
  FILE *inifile;
  struct hostent *hp;
  u32 addr;
  char buf[BUFLEN];

  if ((inifile=fopen(INIFILE, "r")) == NULL)
    return 0; /* no ini file */

  if (fgets(buf, BUFLEN, inifile) == NULL) { /* first choice dest */
    fclose(inifile);
    return 0; /* fault */
  }
  if ((addr = inet_addr(buf)) == -1) {
    if ((hp = gethostbyname(buf)) == NULL)
      addr = -1;
    else 
      addr = *(u32 *)hp->h_addr_list[0];
  }
  if (addr != -1)
    proxy1addr = ntohl(addr); /* proxyaddr is kept in host byte order */

  if (fgets(buf, BUFLEN, inifile) == NULL) { /* second choice dest */
    fclose(inifile);
    return 0; /* fault */
  }
  if ((addr = inet_addr(buf)) == -1) {
    if ((hp = gethostbyname(buf)) == NULL)
      addr = -1;
    else 
      addr = *(u32 *)hp->h_addr_list[0];
  }
  if (addr != -1)
    proxy2addr = ntohl(addr); /* proxyaddr is kept in host byte order */

  if (fgets(buf, BUFLEN, inifile) == NULL) { /* port */
    fclose(inifile);
    return 0; /* fault */
  }
  proxyport = atoi(buf) & 0xffff;
  if (fgets(buf, BUFLEN, inifile) == NULL) { /* mail flag */
    fclose(inifile);
    return 0; /* fault */
  }
  protocol = atoi(buf) & 0xffff;
  
  if (fgets(buf, BUFLEN, inifile) == NULL) { /* sender id */
    fclose(inifile);
    return 0; /* fault */
  }
  buf[strlen(buf)-1] = '\0';
  strncpy(frombuf, buf, 128);
  if (fgets(buf, BUFLEN, inifile) != NULL) { /* freeform - optional */
    buf[strlen(buf)-1] = '\0';    
    strncpy(freeform, buf, 256);
  }
  fclose(inifile);
  return 1;
}


/* sendit - send information by mail or TCP, or just log it */
void
sendit(char *buf, int len)
{
  FILE *handle, *buffile;
  int status;
  int fd, i, fault;
  struct sockaddr_in addr;
  char lbuf[MSGLEN+3];

  fault = 0;
  switch (protocol) {
  case PROTO_MAIL:

    handle = popen("/usr/lib/sendmail " MAILADDR, "w");
    if (handle == NULL) {
      puts("Warning: couldn't pipe to sendmail\n");
      fflush(stdout);
    }
    else {
      fputs(buf, handle);
      fflush(handle);
      status = pclose(handle);
      if (status != 0) {
	printf("Warning: sendmail returned status %d\n", status);
	fflush(stdout);
      }
    }
    break;

  case PROTO_TCP:
    
    fault = 0;
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd > 0) {    
      addr.sin_family = AF_INET;
      addr.sin_port = htons(proxyport);
      addr.sin_addr.s_addr = htonl(proxy1addr);
      if (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) >= 0) {
	if ((buffile=fopen(BUFFILE, "r")) != NULL) {
	  while(fgets(lbuf, MSGLEN+3, buffile) != NULL)
	    if (send(fd, lbuf, strlen(lbuf), 0) < 0) {
	      fault = 1;
	      break;
	    }
	  fclose(buffile);
	  if (!fault)
	    remove(BUFFILE);
	}
	if (send(fd, buf, len, 0) < 0)
	  fault = 1;
      }
      else
	fault = 1;
      close(fd);
    }
    else
      fault = 1;

    if (!fault) 
      break;
      
    fault = 0;
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd > 0) {    
      addr.sin_family = AF_INET;
      addr.sin_port = htons(proxyport);
      addr.sin_addr.s_addr = htonl(proxy2addr);
      if (connect(fd, (struct sockaddr *)&addr, sizeof(addr)) >= 0) {
	if ((buffile=fopen(BUFFILE, "r")) != NULL) {
	  while(fgets(lbuf, MSGLEN+3, buffile) != NULL)
	    if (send(fd, lbuf, strlen(lbuf), 0) < 0) {
	      fault = 1;
	      break;
	    }
	  fclose(buffile);
	  if (!fault)
	    remove(BUFFILE);
	}
	if (send(fd, buf, len, 0) < 0)
	  fault = 1;
      }
      else
	fault = 1;
      close(fd);
    }
    else
      fault = 1;
    break;

  default:
  } /* switch */
  
  if (fault && ((buffile=fopen(BUFFILE, "a")) != NULL)) {
    fputs(buf, buffile);
    fclose(buffile);
  }
  fputs(buf, stdout); /* log it */
  fflush(stdout);
}

/* gettime - get current process usage time */

double
gettime()
{
  struct rusage ru;

  getrusage(RUSAGE_SELF, &ru);
  return ru.ru_utime.tv_sec+ru.ru_stime.tv_sec +
	(ru.ru_utime.tv_usec+ru.ru_stime.tv_usec)/1000000.0;
}



#define PARAL (32UL)

/* Do not change these. */
#define ADD_MASK (15UL)
#define ADDERS (15UL)
#define POW_WRITE (30UL)

int main(int argc, char **argv)
{
	int z1 = 1, z2 = 1;

	u32 i, itersTh[PARAL], itersTl[PARAL];
	int zA[ADDERS], z3T[PARAL], zT[PARAL];
	u128 uA[ADDERS], xA[ADDERS], yA[ADDERS], u3T[PARAL], v3T[PARAL],
		x3T[PARAL], y3T[PARAL], xT[PARAL], yT[PARAL], denT[PARAL],
		numT[PARAL], startvT[PARAL];
	char outbuf[BUFLEN];
	
	/* Timing */
	double starttime, curtime, savetime, pointtime[PARAL];

	/* save state */
	FILE *savfile;

	if (!load_ini()) {
	  printf("No ini file! Run 'timer97a -i' to create one\n");
	  exit(1);
	}

	puts("Getting curve...");
	a = encode(a);
	b = encode(b);

	puts("Getting initial points...");
	x1 = encode(x1);
	y1 = encode(y1);
	x2 = encode(x2);
	y2 = encode(y2);
	{ 
		u32 startu;

		startu = 1UL;
		for (i = 0UL; i < ADDERS; ++i) 
		{
			u128 fac;
			const u32 pi = 3141592653UL; /* pi times 1e9 */
			startu *= pi;
			fac.ex = fac.hi = 0UL; 
			fac.mi = fac.lo = startu;
      
			uA[i] = fac;
			zA[i] = ellipticProduct(x1, y1, z1, fac, a, xA+i, yA+i);
		} 
	}

	/* Reload sav file, if it exists, then initialise any shortfall with
	   random start points */

  i = 0;
  {
    u128 xu, yu, xv, yv;
    u32 zu, zv;

    if ((savfile=fopen(SAVFILE, "r")) != NULL) {
      while (!feof(savfile)) {
	fscanf(savfile, "%1lx%8lx%8lx%8lx %4lx%8lx %1lx%8lx%8lx%8lx %1lx%8lx%8lx%8lx %lf\n",
	      &startvT[i].ex, &startvT[i].hi, &startvT[i].mi, &startvT[i].lo,
	      &itersTh[i], &itersTl[i],
	      &u3T[i].ex, &u3T[i].hi, &u3T[i].mi, &u3T[i].lo,
	      &v3T[i].ex, &v3T[i].hi, &v3T[i].mi, &v3T[i].lo,
	      &pointtime[i]);
	zu = ellipticProduct(x1, y1, z1, u3T[i], a, &xu, &yu);
	zv = ellipticProduct(x2, y2, z2, v3T[i], a, &xv, &yv);
	z3T[i] = ellipticSum(xu, yu, zu, xv, yv, zv, a, x3T+i, y3T+i);
	if (++i >= PARAL)
	  break;
      }
      fclose(savfile);
      remove(SAVFILE);
      printf("Loaded %d saved points\n", i);
    }
  }

  { 
    u32 seed_hi, seed_lo;

    seed_lo = time(NULL);
    seed_hi = getpid();
      
    /* cut out a lot of seed stuff */

    for (; i < PARAL; ++i) { /* continue the load loop */
      u128 startv;
      const u32 eulerGamma_lo = 0x7db0c7a5, eulerGamma_hi = 0x93c467e3; /* gamma times 2^64 */

      u3T[i].ex = u3T[i].hi = u3T[i].mi = u3T[i].lo = 0UL;
      startv.ex = startv.hi = 0UL; 
      startv.mi = seed_hi; startv.lo = seed_lo;
      
      seed_lo += eulerGamma_lo;
      seed_hi += eulerGamma_hi;
      seed_hi += seed_lo < eulerGamma_lo;
      startvT[i] = startv;
      v3T[i] = startv;
      itersTh[i] = itersTl[i] = 0UL;
      pointtime[i] = 0.0;
      z3T[i] = ellipticProduct(x2, y2, z2, startv, a, x3T+i, y3T+i);
    }
  }

  puts("Computing iterations...");
  fflush(stdout);
  
 
  signal(SIGTERM, term_sig); /* catch terminate signal */

  starttime = gettime();
  for (i=0; i < PARAL; ++i) /* adjust point times */
    pointtime[i] = starttime-pointtime[i];

  /** Main loop. **/
  running = 1;
  while (running) {
    for (i = 0UL; i < PARAL; ++i) {
      if (x3T[i].mi>>(32UL-POW_WRITE) == 0UL) {
	u128 x3, y3;
	u32 ptime, len1, len2;

	ptime = (gettime() - pointtime[i] +0.5); /* round & convert to u32 */
	/* Output distinguished point. */

	decode(&x3, &x3T[i]);
	decode(&y3, &y3T[i]);
    
	sprintf(outbuf, "ECCp-97 s %01lx%08lx%08lx%08lx i %04lx%08lx "
		"x %01lx%08lx%08lx%08lx y %01lx%08lx%08lx%08lx z %x "
		"u %01lx%08lx%08lx%08lx v %01lx%08lx%08lx%08lx t %08lx "
		"%s%02d %s ",
		startvT[i].ex, startvT[i].hi, startvT[i].mi, startvT[i].lo,
		itersTh[i], itersTl[i],	
		x3.ex, x3.hi, x3.mi, x3.lo, y3.ex, y3.hi, y3.mi, y3.lo, z3T[i],
		u3T[i].ex, u3T[i].hi, u3T[i].mi, u3T[i].lo,
		v3T[i].ex, v3T[i].hi, v3T[i].mi, v3T[i].lo, ptime,
		CLIENT, PARAL, VERSION);
	len1 = strlen(outbuf);
	len2 = strlen(frombuf);
	if ((len1+len2) > MSGLEN) {
	  len2 = MSGLEN-len1;
	  frombuf[len2] = '\0'; /* truncate it */
	}
	strcpy(outbuf+len1, frombuf);
	len1 += len2;
	if (len1+2 < MSGLEN) {
	  len2 = strlen(freeform);
	  if ((len1+len2+2) > MSGLEN) {
	    len2 = MSGLEN-2-len1;
	    freeform[len2] = '\0'; /* truncate it */
	  }
	  strcpy(outbuf+len1, " ;");
	  len1 += 2;
	  strcpy(outbuf+len1, freeform);
	  len1 += len2;
	  strcpy(outbuf+len1++, "\n");
	}
	sendit(outbuf, len1);
	
        
	/* Restart this one by setting u to 0... */
	u3T[i].ex = u3T[i].hi = u3T[i].mi = u3T[i].lo = 0UL;
	startvT[i] = v3T[i];
	itersTh[i] = itersTl[i] = 0UL;

	z3T[i] = ellipticProduct(x2, y2, z2, startvT[i], a, x3T+i, y3T+i);
	pointtime[i] = gettime();

      } /* end if (distinguished point) */
    } /* end for (i) */

    
    /* Get numT[] & denT[], update u3T[] & v3T[]. */
    for (i = 0UL; i < PARAL; ++i) {
      u32 m;

      /* Pseudo-random function.  Same on all machines. */ 
      m = x3T[i].lo & ADD_MASK;
      if (m < ADDERS) {
	u3T[i] = addMod(u3T[i], uA[m], ord);
	xT[i] = xA[m]; yT[i] = yA[m]; zT[i] = zA[m];
      } 
      else {
	u3T[i] = doubleMod(u3T[i], ord);
	v3T[i] = doubleMod(v3T[i], ord);
	xT[i] = x3T[i]; yT[i] = y3T[i]; zT[i] = z3T[i];
      } /* end if/else */
		  
      if (equal(&xT[i], &x3T[i])) {
	if (equal(&yT[i], &y3T[i])) { /* Doubling, */
	  u128 s;
            
	  s = squareMod(&xT[i]);
	  numT[i] = doubleMod(s, mod);
	  numT[i] = addMod(numT[i], s, mod);
	  numT[i] = addMod(numT[i], a, mod);
	  denT[i] = doubleMod(yT[i], mod);
	} 
	else {
	  numT[i] = denT[i] = one; /* dummy values */
	}
      } 
      else { /* General case. */
	numT[i] = subMod(yT[i], y3T[i], mod);
	denT[i] = subMod(xT[i], x3T[i], mod);
      } 
    } /* end for (i) */

    /* Divide numT[] by denT[] with one inversion and 4*PARAL-3 mults. */
    { 
      u128 ix, prod, q, prodT[PARAL];
  
      prod = denT[0];   
      for (i = 1UL; i < PARAL; ++i) {
	prodT[i] = prod;
	prod = productMod(&prod, &denT[i]);
      } /* end for */
      q = inverseMod(prod);
      for (i = PARAL; --i; ) {
	ix = productMod(&q, &prodT[i]);
	numT[i] = productMod(&numT[i], &ix);
	q = productMod(&q, &denT[i]);
      } /* end for */
     
      numT[0] = productMod(&numT[0], &q);
    } /* end block */
   
    /* Get new points. */
    
    for (i = 0UL; i < PARAL; ++i) {
      u128 nx, ny;

      if (equal(&xT[i], &x3T[i])) {
	const u128 zero = {0UL, 0UL, 0UL, 0UL};
		  
	if (equal(&yT[i], &y3T[i])) { /* Doubling, */  
	  u128 lam, s, t;
			  
	  lam = numT[i];
			  
	  nx = squareMod(&lam);
	  s = doubleMod(xT[i], mod);
	  nx = subMod(nx, s, mod);

	  t = subMod(xT[i], nx, mod);
	  ny = productMod(&lam, &t);
	  ny = subMod(ny, yT[i], mod);

	  z3T[i] = 1;
	} 
	else { 
	  nx.ex = nx.hi = nx.mi = nx.lo = z3T[i] = 0; 
	  ny = one;
	}
      
      } 
      else { /* General case. */
	u128 lam, s;

	lam = numT[i];

	nx = squareMod(&lam);
	nx = subMod(nx, xT[i], mod);
	nx = subMod(nx, x3T[i], mod);

	s = subMod(xT[i], nx, mod);
	ny = productMod(&lam, &s);
	ny = subMod(ny, yT[i], mod);

	z3T[i] = 1;  
      } /* end if/else/else... */

      x3T[i] = nx; y3T[i] = ny; 
      ++itersTl[i];
      if (itersTl[i] == 0)
        ++itersTh[i];

    } /* end for (i) */

  } /* end forever */

	
  /* save state */

  savetime = gettime();
  if ((savfile=fopen(SAVFILE, "w")) != NULL) {
    for (i = 0; i < PARAL; i++)
      fprintf(savfile, "%01lx%08lx%08lx%08lx %04lx%08lx %01lx%08lx%08lx%08lx "
	      "%01lx%08lx%08lx%08lx %12.4f\n",
	      startvT[i].ex, startvT[i].hi, startvT[i].mi, startvT[i].lo,
	      itersTh[i], itersTl[i],
	      u3T[i].ex, u3T[i].hi, u3T[i].mi, u3T[i].lo,
	      v3T[i].ex, v3T[i].hi, v3T[i].mi, v3T[i].lo,
	      (savetime-pointtime[i]));
    fclose(savfile);
  }

	return 0;

}

